#!/usr/bin/env python3
'''Read data from WEB via `requests` [GET|POST] method
Usage of,
./webReader.py URL [JSON]
OR
./webReader.py AAPL 'dict(funcName="yhReader")'

Where:
JSON default value:
{ "funcName":"webReader",
  "opts": { "rtype":"json", "method":"get" }
}

Example,
#----- To read web data from FRED
./webReader.py 'https://api.stlouisfed.org/fred/series/observations?file_type=json&series_id=CPIAUCNS' $(cat ~/.fredapi.json)
#----- OR 
api_key=$(jq ".api_key" ~/.fredapi.json)
./webReader.py 'https://api.stlouisfed.org/fred/series/observations?' '{"api_key":'\"${api_key}\"',"file_type":"json","series_id":"CPIAUCNS","observation_start":"2022-07-01"}'
#----- OR to read web data from yahoo-finance
./webReader.py 'https://query2.finance.yahoo.com/v7/finance/spark?range=1d&interval=30m&indicators=close&includeTimestamps=false&includePrePost=false&corsDomain=finance.yahoo.com&.tsrc=finance' '{"symbols":"AAPL,IBM"}'
#----- OR 
./webReader.py 'https://query2.finance.yahoo.com/v7/finance/spark' '{"symbols":"AAPL","range":"1d","interval":"60m"}'
#----- OR to load a streaming file 
./webReader.py http://api1.beyondbond.com/downloads/grimm10-49_1.png {} stream get "${user},${pswd}"
#----- OR to load stock spark prices 
./webReader.py AAPL,TSM 'dict(funcName="yhReader")'
#----- OR 
python3 -c "from webReader import yhReader;df=yhReader(['AAPL','AMD'],types='spark');print(df)"
#----- OR 
python3 -c "from webReader import yhReader;df=yhReader(['AAPL','AMD'],types='chart');print(df)"
#----- OR 
python3 -c "from webReader import yhReader;df=yhReader(['AAPL','AMD'],types='quote');print(df)"

Last Mod.,
Sun 02 Jul 2023 06:52:03 PM EDT
'''
__usage__='\n'.join(__doc__.split('\n')[:11])
import sys
import requests
from requests.auth import HTTPBasicAuth as HA
import pandas as pd

headers={'Content-Type': 'text/html', 'Accept': 'application/json', 'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/59.0.3071.115 Safari/537.36'}

credentials={}

apiBase = 'https://query2.finance.yahoo.com'

def getCredentials(cookieUrl='https://fc.yahoo.com', crumbUrl=apiBase+'/v1/test/getcrumb'):
	cookies = requests.get(cookieUrl).cookies
	crumb = requests.get(url=crumbUrl, cookies=cookies, headers=headers).text
	return {'cookies': cookies, 'crumb': crumb}

def chart_json2df(res):
	jx=res["chart"]["result"][0]
	ticker=jx["meta"]["symbol"]
	epoch=jx["timestamp"]
	dlst=jx["indicators"]["quote"][0]
	dlst.update(epoch=jx["timestamp"])
	if "adjclose" in jx["indicators"]:
		adjusted=jx["indicators"]["adjclose"][0]["adjclose"]
	else:
		adjusted=dlst["close"]
	dlst.update(epoch=epoch,adjusted=adjusted)
	dx=pd.DataFrame(data=dlst)
	dx['ticker']=ticker
	dx.set_index(pd.DatetimeIndex(dx['epoch'].apply(pd.Timestamp.fromtimestamp)),inplace=True)
	return dx

def spark_json2df(res):
	jlst=res["spark"]["result"]
	df=pd.DataFrame()	
	for jx in jlst:
		epoch=jx["response"][0]["timestamp"]
		close=jx["response"][0]["indicators"]["quote"][0]["close"]
		ticker=jx["symbol"]
		dx=pd.DataFrame(data=dict(epoch=epoch,close=close,ticker=ticker))
		dx.set_index(pd.DatetimeIndex(dx['epoch'].apply(pd.Timestamp.fromtimestamp)),inplace=True)
		df = pd.concat([df,dx])
	df.index.name="DATE"
	df=df.drop(["epoch"],axis=1)
	df["pbdate"]=pd.Series((df.index)).apply(lambda x: pd.Timestamp.strftime(x,"%Y%m%d")).values
	return df

def yh_chart(tkLst=['AAPL'],debugTF=False,**opts):
	if debugTF:
		sys.stderr.write("==yh_chart inputs:{}\n".format(locals()))
	urx="https://query2.finance.yahoo.com/v8/finance/chart/{}"
	parat=dict(interval='1d',range='7d')
	pLst=['interval','range','fields','period1','period2','includeAdjustedClose']
	parax={k:v for k,v in opts.items() if k in pLst}
	parat.update(parax)
	df=pd.DataFrame()	
	for tx in tkLst:
		url=urx.format(tx)
		res=webReader(url,pjson=parat,debugTF=debugTF,**opts)
		dx = chart_json2df(res)
		df = pd.concat([df,dx])
	df.index.name="DATE"
	df=df.drop(["epoch"],axis=1)
	df["pbdate"]=pd.Series((df.index)).apply(lambda x: pd.Timestamp.strftime(x,"%Y%m%d")).values
	if parat.get('interval','')[-1:]=="m":
		df["epochs"]=pd.Series((df.index)).apply(lambda x: pd.Timestamp.strftime(x,"%s000")).values
	if not opts.get('dfTF',False):
		return df.to_dict(orient='records')
	return df

def yh_spark(tkLst=['AAPL'],debugTF=False,**opts):
	if debugTF:
		sys.stderr.write("==yh_spark inputs:{}\n".format(locals()))
	url="https://query2.finance.yahoo.com/v7/finance/spark?"
	parat=dict(symbols=",".join(tkLst),interval='1d',range='7d')
	pLst=['symbols','interval','range','fields','indicators','period1','period2']
	parax={k:v for k,v in opts.items() if k in pLst}
	parat.update(parax)
	res=webReader(url,pjson=parat,debugTF=debugTF,**opts)
	df=spark_json2df(res)
	if parat.get('interval','')[-1:]=="m":
		df["epochs"]=pd.Series((df.index)).apply(lambda x: pd.Timestamp.strftime(x,"%s000")).values
	return df

def yh_quote(tkLst=['AAPL'],debugTF=False,**opts):
	if debugTF:
		sys.stderr.write("==yh_quote inputs:{}\n".format(locals()))
	url="https://query2.finance.yahoo.com/v7/finance/quote?"
	parat=dict(symbols=",".join(tkLst))
	pLst=['symbols','fields','formatted']
	parax={k:v for k,v in opts.items() if k in pLst}
	parat.update(parax)
	res=webReader(url,pjson=parat,debugTF=debugTF,**opts)
	jlst=res["quoteResponse"]["result"]
	if opts.get('dfTF',False):
		return pd.DataFrame(jlst)
	else:
		return jlst

def yhReader(tkLst,types='spark',debugTF=False,**opts):
	if isinstance(tkLst,str):
		tkLst=tkLst.split(',')
	typ=types.lower()
	if typ =='chart':
		return yh_chart(tkLst=tkLst,debugTF=debugTF,**opts)
	elif typ =='spark':
		return yh_spark(tkLst=tkLst,debugTF=debugTF,**opts)
	elif typ =='quote':
		return yh_quote(tkLst=tkLst,debugTF=debugTF,**opts)
	return {}

def webReader(url,pjson={},rtype='json',method='get',sessionTF=True,debugTF=False,**opts):
	global credentials
	if not url:
		return {}
	user, pswd = [opts.pop(x,'') for x in ['user','pswd'] ]
	if debugTF:
		sys.stderr.write("==webReader inputs:{}\n".format(locals()))
	rq = requests.Session() if sessionTF else requests
	rqOpts={}
	if 'quote' in url and 'quoteS' not in url:
		if not credentials:
			credentials=getCredentials()
		crumb,cookies = (credentials['crumb'],credentials['cookies'])
		pjson.update(crumb=crumb)
		rqOpts.update(cookies=cookies)
	if (user and pswd):
		rqOpts.update(auth=HA(user,pswd))
	if debugTF:
		sys.stderr.write(" --requests URL:{}\n --params:{}\n --rqOpts:{}\n".format(url,pjson,rqOpts))
	if method.lower()=='post':
		ret = rq.post(url,data=pjson,headers=headers,**rqOpts)
	else:
		ret = rq.get(url,params=pjson,headers=headers,**rqOpts)
	res = ret.json() if rtype.lower()=='json' else ret.text
	return res
	
def streamReader(url,pjson={},rtype='json',method='get',sessionTF=True,debugTF=False,**opts):
	from urllib.parse import urlparse
	import shutil
	if not url:
		return {}
	stream = opts.pop('stream',True)
	user, pswd,filepath = [opts.pop(x,'') for x in ['user','pswd','filepath'] ]
	if debugTF:
		sys.stderr.write("==streamReader inputs:{}\n".format(locals()))
	if not filepath:
		o=urlparse(url)
		filepath=o.path.split('/')[-1]
	rq = requests.Session() if sessionTF else requests
	rqOpts={}
	if stream:
		rqOpts.update(stream=stream)
	if (user and pswd):
		rqOpts.update(auth=HA(user,pswd))
	if debugTF:
		sys.stderr.write(" --requests URL:{}\n --params:{}\n --rqOpts:{}\n".format(url,pjson,rqOpts))
	if method.lower()=='post':
		ret = rq.post(url,data=pjson,headers=headers,**rqOpts)
	else:
		ret = rq.get(url,params=pjson,headers=headers,**rqOpts)
	if ret.status_code == 200:
		with open(filepath, 'wb') as f:
			ret.raw.decode_content = True
			shutil.copyfileobj(ret.raw, f)  
	else:
		print(ret.status_code,ret.raise_for_status(),file=sys.stderr)
	del ret
	return filepath
	
def mainTst():
	pjson={}
	argc=len(sys.argv)
	if argc>1:
		url = sys.argv[1]
	else:
		sys.exit(__usage__)
	try:
		pjson=eval(sys.argv[2]) if sys.argv[2:] else {}
	except Exception as e:
		sys.stderr.write("**ERROR: Invalid JSON, {}\n".format(e))
	opts = pjson.pop('opts',{})
	funcName = pjson.pop('funcName','webReader')
	if funcName in globals() and hasattr(globals()[funcName],'__call__'):
		funcArg=globals()[funcName]
	else:
		sys.stderr.write("**WARNINGS: Invalid function:{}\n".format(funcName))
		sys.exit(__usage__)

	if funcArg in [streamReader,webReader]:
		res=funcArg(url,pjson=pjson,**opts)
		if 'spark' in url:
			res = spark_json2df(res)
		elif 'chart' in url:
			res = chart_json2df(res)
	else:
		if opts:
			pjson.update(opts)
		res=funcArg(url,**pjson)
	return res

if __name__ == '__main__' :
	try:
		sys.exit(mainTst())
	except Exception as e:
		sys.stderr.write("**ERROR: mainTst {}\n".format(e))
		sys.exit(__usage__)




#!/usr/bin/env python3
'''Fetch REQUEST INFO
'''
import sys,os

def evv(va):
	try:
		return eval(va)
	except:
		return va

def qsparse(qs='',**opts):
	import urllib.parse
	if not qs:
		qs=os.getenv('QUERY_STRING')
	if not qs:
		return opts
	qs=urllib.parse.unquote_plus(qs)
	for xs in qs.strip().split('&'):
		vs=xs.strip().split('=')
		if len(vs)==2:
			opts.update({vs[0]:evv(vs[1])})
		elif len(vs)>2:
			opts.update({vs[0]:evv("=".join(vs[1:]))})
	return opts

def rqparse(xreq='POST', debugTF=False):
	if xreq=='POST':
		pst=sys.stdin.read().strip()
	else:
		pst=''
	opts=qsparse() #DEPRECATED, opts=get_cgi(xreq)
	if pst:
		opts.update(eval(pst))
	if not debugTF:
		return opts
	remoteAddr=os.getenv('REMOTE_ADDR')
	if remoteAddr:
		opts.update(remoteAddr=remoteAddr)
	realAddr=os.getenv('HTTP_X_REAL_IP')
	if realAddr:
		opts.update(remoteAddr=realAddr)
	envLst=",".join(["({}:{})".format(x,y) for x,y in os.environ.items()])
	opts.update(envLst=envLst)
	return opts

def get_rq(**optx):
	xreq=os.getenv('REQUEST_METHOD')
	if xreq not in ('GET','POST'):
		return optx
	try:
		optx=rqparse(xreq)
	except Exception as e:
		optx={"err":str(e)}
	ctype=optx.get('ctype','plain')
	if ctype=='html':
		print("Content-type:text/html;charset=utf-8\r\n\r\n")
	elif ctype=='json':
		print("Content-type:application/json;charset=utf-8\r\n\r\n")
	else:
		print("Content-type:text/plain;charset=utf-8\r\n\r\n")
	return optx

def prn_df(df,**optx):
	ctype=optx.get('ctype','plain')
	indexTF=optx.get('indexTF',False)
	# use commas to separate thousands in floats, complex numbers, and integers
	clx=df.columns
	cfm={s:'{:,}'.format for s in clx if df[s].values.dtype in ['int64']}
	cfx={s:'{:,.2f}'.format for s in clx if df[s].values.dtype in ['float64']}
	cfm.update(cfx)
	_ = cfm.pop('pbdate',None)
	cfx=optx.get('cfm',{})
	if cfx:
		cfm.update(cfx)
	if ctype=='html':
		print(df.to_html(index=indexTF,justify='right',formatters=cfm))
	elif ctype=='json':
		print(df.to_json(orient='records'))
	elif ctype=='dictionary': # special debugging format for `dict` type
		from pprint import pprint
		pprint(df.to_dict(orient='records'))
	else:
		print(df.to_string(index=indexTF,formatters=cfm))



#!/usr/bin/env python3
'''chatGPT API Quickstart
Usage of,
./chatgtpAPI.py JSON
/cgi-bin/chatgptapi?model=MODEL&prompt=PROMPT

#--From commandline
chatgptAPI.py '{"prompt":"plan me a 2-day aruba trip"}'

Example, 
#--From internet
/cgi-bin/chatgptapi?prompt=suggest 3 names of my cat
OR
/cgi-bin/chatgptapi?model=text-davinci-edit-001&input=when%20my%20birthday&instruction=fix%20sentence

List of model:
"text-davinci-003": # create completions
"text-davinci-edit-001": # edits
"text-embedding-ada-002": # create embddings
"images-generations": # create images
"images-edits": # edit images
"images-variations": # images variation
"list-files": # list files
"upload-files": # upload files
"delete-files": # delete files

Last Mod.,
Tue 17 Jan 2023 05:35:49 PM EST
'''
import sys,os
import cgi,cgitb
import json
import datetime

def evv(va):
	try:
		return eval(va)
	except:
		return va
		
def get_cgi(mth,**opts):
	import cgi,cgitb
	cgitb.enable(logdir="/apps/fafa/bb_site/cgi-bin/.log")
	cgitb.enable(format='text',display=False)
	mf = cgi.FieldStorage()
	sys.stdout.write(" --{}-Args: {}\n".format(mth,mf))
	for ky in mf:
		va=mf[ky].value
		if ky[-2:] == 'TF':
			va = True if va[:1].lower() in ['t','y','1'] else False
		else:
			va = evv(va)
		opts.update({ky:va})
	return opts

def ulopen(url,rtype="rb"):
	import urllib.request
	if url[:4].lower()=='http':
		req=urllib.request.Request(url)
		return urllib.request.urlopen(req)
	else:
		return open(url,rtype)

def chatgptAPI(**opts):
	import requests
	import json

	rqOpts = opts.pop("rqOpts",{})
	pathname=opts.pop('pathname',None)
	debugTF=opts.pop('debugTF',False)
	xtra=evv(opts.pop('xtra',''))
	if isinstance(xtra,dict):
		opts.update(xtra)

	hds={'Content-Type': 'application/json', 'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/59.0.3071.115 Safari/537.36'}
	if not any([True for x in ["image","prompt","input","instruction"] if x in opts]):
		return dict(err="**WARNINGS:required `prompt` in opts:{}".format(opts))
	model = opts.get("model","text-davinci-003")
	if debugTF:
		sys.stderr.write("INPUT opts:{}\n".format(opts))
	if model in ["text-davinci-003","text"]: # create completions
		if not any([True for x in ["prompt"] if x in opts]):
			return {}
		url = opts.pop("url","https://api.openai.com/v1/completions")
		data={"model": model,"top_p": 0.1,"max_tokens": 4000}
	elif model=="text-davinci-edit-001": # edits
		if 'instruction' not in opts: 
			opts.update(instruction="correct it")
		if 'prompt' in opts: 
			input=opts.pop('prompt','')
			if 'input' not in opts:
				opts.update(input=input)
		url = opts.pop("url","https://api.openai.com/v1/edits")
		data={"model": model,"top_p": 0.1}
	elif model=="text-embedding-ada-002": # create embddings
		if 'prompt' in opts: 
			input=opts.pop('prompt','')
			if 'input' not in opts:
				opts.update(input=input)
		if not any([True for x in ["input"] if x in opts]):
			return dict(err="**WARNINGS:[{}] required `input` in opts:{}".format(model,opts))
		url = opts.pop("url","https://api.openai.com/v1/embeddings")
		data={"model": model,"top_p": 0.1}
	elif model in ["images-generations","image","images","create-image","create-images"]: # create images
		if not any([True for x in ["prompt"] if x in opts]):
			return {}
		url = opts.pop("url","https://api.openai.com/v1/images/generations")
		data={"n":1,"size": "512x512"}
	elif model in ["images-edits","edit-images","edit-image"]: # edit images
		if not all([True for x in ["image","prompt"] if x in opts]):
			return {}
		url = opts.pop("url","https://api.openai.com/v1/images/edits")
		data={"n":1,"size": "512x512"}
		#hds.update({'Content-Type': 'multipart/form-data'})
		hds.pop('Content-Type',None)
		files=dict(image=ulopen(opts.pop('image'),"rb"))
		rqOpts.update(files=files)
	elif model=="images-variations": # images variation
		if not all([True for x in ["image"] if x in opts]):
			return {}
		if 'prompt' in opts: 
			image=opts.pop('prompt','')
			if 'image' not in opts:
				opts.update(image=image)
		url = opts.pop("url","https://api.openai.com/v1/images/variations")
		data={"n":1,"size": "512x512"}
		#hds.update({'Content-Type': 'multipart/form-data'})
		hds.pop('Content-Type',None)
		files=dict(image=ulopen(opts.pop('image'),"rb"))
		rqOpts.update(files=files)
	elif model=="list-files": # list files
		url = opts.pop("url","https://api.openai.com/v1/files")
		data={}
	elif model=="upload-files": # upload files
		if not all([True for x in ["file","purpose"] if x in opts]):
			return {}
		url = opts.pop("url","https://api.openai.com/v1/files")
		data={}
	elif model=="delete-files": # delete files
		if not all([True for x in ["file_id"] if x in opts]):
			return {}
		url = opts.pop("url","https://api.openai.com/v1/files")
		url = "{}/{}".format(url,opts["file_id"])
		data={"object":"file","deleted":True}
	else:
		return {}
		
	headers = opts.pop("headers",hds)
	sessionTF = opts.pop("sessionTF",False)
	if "OPENAI_API_KEY" in os.environ:
		auth='Bearer {}'.format(os.environ["OPENAI_API_KEY"])
	else:
		auth='Bearer {}'.format(open('/home/rstudio/.OPENAI_API_KEY','r').read().strip())
	headers.update(Authorization=auth) 
	
	if opts:
		for x,y in opts.items():
			if x not in ['model','pathname','submit','xtra']:
				data[x]=evv(y)

		if "temperature" in data:
			data.pop("top_p",None)
			
	rq = requests.Session() if sessionTF else requests
	if debugTF:
		sys.stderr.write("+++url:{}\ndata:{}\nheaders:{}\nrqOpts:{}".format(url,data,headers,rqOpts))
	if model=="delete-files": # delete files
		ret = rq.delete(url,headers=headers,**rqOpts)
	else:
		if 'Content-Type' in headers and headers['Content-Type']=='application/json':
			ret = rq.post(url,data=json.dumps(data),headers=headers,**rqOpts)
		else:
			ret = rq.post(url,data=data,headers=headers,**rqOpts)
	res = ret.json()
	data.update(model=model)
	res['input_param']=data
	return res
	
if __name__ == '__main__' :
	xreq=os.getenv('REQUEST_METHOD')
	if xreq in ('GET','POST'):
		pst=sys.stdin.read()
		optx=get_cgi(xreq)
		try:
			if pst:
				optx.update(eval(pst))
		except Exception as e:
			optx.update(err=str(e))
		ctype=optx.pop('ctype','json')
		if ctype=='html':
			print("Content-type:text/html;charset=utf-8\r\n\r\n")
		elif ctype=='plain':
			print("Content-type:text/plain;charset=utf-8\r\n\r\n")
		elif ctype=='multipart':
			print("Content-type:multipart/form-data\r\n\r\n")
		else:
			print("Content-type:application/json;charset=utf-8\r\n\r\n")
	else:
		optx=eval(sys.argv[1]) if sys.argv[1:] else {}
		ctype=optx.pop('ctype','json')
	debugTF=optx.get('debugTF',False)
	if not optx:
		if xreq in ('GET','POST'):
			print(json.dumps(dict(err=__doc__)) if ctype=='json' else __doc__)
		sys.exit(__doc__)
	if debugTF:
		sys.stderr.write("optx:{}\n".format(optx))
	res=chatgptAPI(**optx)

	if ctype=='json':
		if debugTF:
			print(json.dumps([optx,res]))
		else:
			print(json.dumps(res))
		try:
			remoteAddr=os.getenv('HTTP_X_REAL_IP')
			if not remoteAddr:
				remoteAddr=os.getenv('REMOTE_ADDR')
			pbdt=datetime.datetime.now()
			fp=open("/apps/fafa/bb_site/cgi-bin/.log/chatgptapi.log","a")
			fp.write("++ from {} @ {}\n{}\n".format(remoteAddr,pbdt,json.dumps(res)))
		except Exception as e:
			sys.stderr.write("**ERROR on logfile:{}\n".format(e))
	else:
		if 'choices' in res and 'text' in res['choices'][0]:
			print(res['choices'][0]['text'])
		elif 'data' in res and 'url' in res['data'][0]:
			print(res['data'][0]['url'])
		else:
			print("N/A")
